//  Copyright © 2012 Stephen Smally
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.
//      

using PackageKit;
using SQLHeavy;

namespace AppStore { 
    public class AppsManager : Object {
        // Signals
        public signal void app_added (ModelApp app);
        public signal void category_added (Category cat);
        public signal void loading_started (LoadingType load, string text);
        public signal void loading_finished (LoadingType load);
        public signal void loading_progress (int percentage);
        public signal void update_received (ModelApp app);
        public signal void details_received (App app);
        public signal void repo_received (RepoDetail repo);
        public signal void error_received (ErrorType type);
        
        // PackageKit stuffs
        private Client client;
        private Control control;
        private Task task;
        private Cancellable transaction;
        
        // Database stuffs
        private static DBManager? database = null;
        private static string? database_path = null;
        
        public void remove_package (string id, bool deps, bool auto) {
            loading_started (LoadingType.REMOVE, "Removing packages...");
            string[] ids = { id, null };
            task.remove_packages_async.begin  (ids,
                                        deps,
                                        auto,
                                        null,
                                        (p) => {
                                            if (p.percentage != -1) {
                                                loading_progress (p.percentage);
                                            }
                                        },
                                        (obj, res) => {
                                            // generic_finish does not end if transaction was cancelled
                                            try {
                                                var result = task.generic_finish (res);
                                                if (result != null)
                                                    loading_finished (LoadingType.REMOVE);
                                            } catch (GLib.Error e) {
                                                message_log (e.message);
                                                parse_error (e);
                                                loading_finished (LoadingType.CANCEL);
                                            }
                                        });
        }
        
        public void install_package (string id) {
            loading_started (LoadingType.INSTALL, "Installing packages...");
            string[] ids = { id, null };
            task.install_packages_async.begin  (ids,
                                         null,
                                         (p) => {
                                             if (p.percentage != -1) {
                                                 loading_progress (p.percentage);
                                             }
                                         },
                                         (obj, res) => {
                                            // generic_finish does not end if transaction was cancelled
                                            try {
                                                var result = task.generic_finish (res);
                                                if (result != null)
                                                    loading_finished (LoadingType.INSTALL);
                                            } catch (GLib.Error e) {
                                                message_log (e.message);
                                                parse_error (e);
                                                loading_finished (LoadingType.CANCEL);
                                            }
                                         });
        }
        
        public void get_apps (string category) {
            try {
                QueryResult result = database.get_apps_for_category (category);
                ModelApp app;
                for (int record = 1; ! result.finished; record++, result.next ()) {
                    app = new ModelApp ();
                    app.id = result.fetch_string (0);
                    app.name = result.fetch_string (1);
                    app.description = result.fetch_string (2);
                    app.icon = result.fetch_string (3);
                    app_added (app);
                    //if ("x" in result.fetch_string (0))
                        //stdout.printf ("\n%s", result.fetch_string (0));
                }
            } catch (GLib.Error e) {
                message_log (e.message);
            }
        }
        
        public void get_repos () {
            loading_started (LoadingType.DETAILS, "Loading infos...");
            client.get_repo_list_async.begin (common_filter, null,
                                        () => {
#if HAVE_GTK
                                          while (Gtk.events_pending()) {
                                              Gtk.main_iteration();
                                          }
#endif
                                        },
                                        (obj, res) => {
                                          try {
                                              Results result = client.generic_finish (res);
                                              result.get_repo_detail_array ().foreach ((repo) => {
                                                if (repo != null)
                                                    repo_received (repo);
                                              });
                                          } catch (GLib.Error e) {
                                              message_log (e.message);
                                              parse_error (e);
                                          }
                                          loading_finished (LoadingType.REPOS);
                                        });
        }
        
        private void get_details_2 (ModelApp mapp, Package p) {
            client.get_details_async.begin ({p.get_id(), null},
                                      null,
                                      () => {
#if HAVE_GTK
                                          while (Gtk.events_pending()) {
                                              Gtk.main_iteration();
                                          }
#endif
                                      },
                                      (obj, result) => {
                                          try {
                                              Results res = client.generic_finish (result);
                                              App app = new App (mapp.id,
                                                                  mapp.name,
                                                                  mapp.description,
                                                                  mapp.icon,
                                                                  p,
                                                                  res.get_details_array()[0]);
                                              details_received (app);
                                          } catch (GLib.Error e) {
                                              message_log (e.message);
                                              parse_error (e);
                                          }
                                          loading_finished (LoadingType.DETAILS);
                                      }
                                      );
        }
        
        public void get_details (ModelApp app) {
            loading_started (LoadingType.DETAILS, "Loading infos...");
            
            /*var db = new Appstream.Database ();
            db.database_path = "/var/cache/apt-xapian-index/index.2/";
            db.open ();
            
            var query = new Appstream.SearchQuery ("gnome-shell");
            var apps = db.find_applications (query);
            
            debug (apps.index (0).to_string ());*/
            client.resolve_async.begin  (common_filter,
                                { app.id, null },
                                null,
                                () => {
#if HAVE_GTK
                                    while (Gtk.events_pending()) {
                                        Gtk.main_iteration();
                                    }
#endif
                                },
                                (obj, resu) => {
                                    try {
                                        Results res = client.generic_finish (resu);
                                        get_details_2 (app, res.get_package_array()[0]);
                                    } catch (GLib.Error e) {
                                        message_log (e.message);
                                        parse_error (e);
                                    }
                                });
        }
        
        private int count_category (string category) {
            try {
                return database.get_rows_for_category (category);
            } catch (GLib.Error e) {
                message_log (e.message);
                parse_error (e);
            }
            return -1;
        }
        
        public void get_categories () {
            loading_started (LoadingType.CATEGORIES, "Loading categories...");
            try {
                QueryResult result = database.get_categories ();
                
                for (int record = 1; ! result.finished; record++, result.next ()) {
                    category_added (new Category (result.fetch_string (0),
                                                     result.fetch_string (1),
                                                     result.fetch_string (2),
                                                     result.fetch_string (3),
                                                     count_category (result.fetch_string (0))));
                }
            } catch (GLib.Error e) {
                message_log (e.message);
                parse_error (e);
            }
            
            loading_finished (LoadingType.CATEGORIES);
        }
        
        public void search_for_apps_global (string search_string) {
            // Categories
            try {
                // Check apps for them name, ids and description
                QueryResult result = database.get_categories ();
                
                for (int record = 1; ! result.finished; record++, result.next ())
                    search_for_apps (result.fetch_string(0), search_string.down ());
                    
            } catch (GLib.Error e) {
                message_log (e.message);
                parse_error (e);
            }
        }
          
        private void search_for_apps (string category, string search_string) {
            try {
                QueryResult result = database.get_apps_for_category (category);
                ModelApp app;
                
                for (int record = 1; ! result.finished; record++, result.next ()) {
                    // Look for apps from them ids, names and descriptions
                    if (search_string in result.fetch_string (0).down () ||
                        search_string in result.fetch_string (1).down () ||
                        search_string in result.fetch_string (2).down ()) { // If something is found let's send the application
                        app = new ModelApp ();
                        app.id = result.fetch_string (0);
                        app.name = result.fetch_string (1);
                        app.description = result.fetch_string (2);
                        app.icon = result.fetch_string (3);
                        app_added (app);
                    }
                }
            } catch (GLib.Error e) {
                message_log (e.message);
                parse_error (e);
            }
        }
        
        public bool load_app_from_id (string id) {
            // Categories
            try {
                // Check apps for them name and ids
                QueryResult result = database.get_categories ();
                
                for (int record = 1; ! result.finished; record++, result.next ()) {
                    var category = result.fetch_string(0);
                    
                    try {
                        QueryResult res = database.get_apps_for_category (category);
                        ModelApp app;
                        
                        for (int rec = 1; ! res.finished; rec++, res.next ()) {
                            // Look for apps from them ids and names
                            if (id == res.fetch_string (0)) { // If something is found let's send the application
                                app = new ModelApp ();
                                app.id = res.fetch_string (0);
                                app.name = res.fetch_string (1);
                                app.description = res.fetch_string (2);
                                app.icon = res.fetch_string (3);
                                get_details (app);
                                return true;
                            }
                        }
                    } catch (GLib.Error e) {
                        message_log (e.message);
                        parse_error (e);
                    }
                }
                    
            } catch (GLib.Error e) {
                message_log (e.message);
                parse_error (e);
            }
            
            return false;
        }
        
        private void load_update (Package pkg) {
            // Categories
            try {
                // Look for app for its name
                QueryResult result = database.get_categories ();
                
                for (int record = 1; ! result.finished; record++, result.next ()) {
                    var category = result.fetch_string(0);
                    
                    try {
                        QueryResult res = database.get_apps_for_category (category);
                        ModelApp app;
                        
                        for (int rec = 1; ! res.finished; rec++, res.next ()) {
                            if (pkg.get_name () == res.fetch_string (0)) { // If something is found let's send the application
                                app = new ModelApp ();
                                app.id = res.fetch_string (0);
                                app.name = res.fetch_string (1);
                                app.description = res.fetch_string (2);
                                app.icon = res.fetch_string (3);
                                update_received (app);
                            }
                        }
                    } catch (GLib.Error e) {
                        message_log (e.message);
                        parse_error (e);
                    }
                }
                    
            } catch (GLib.Error e) {
                message_log (e.message);
                parse_error (e);
            }
        }
        
        public void get_updates () {
            loading_started (LoadingType.UPDATES, "Loading updates...");
            task.get_updates_async.begin (common_filter, transaction, () => {         
#if HAVE_GTK
                                    while (Gtk.events_pending()) {
                                        Gtk.main_iteration();
                                    }
#endif
                                    },
                                    (obj, res) => {
                                        try {
                                            Results result = task.generic_finish (res);
                                            result.get_package_array ().foreach ((pkg) => { load_update (pkg); });
                                        } catch (GLib.Error e) {
                                            message_log (e.message);
                                        }
                                    });
            loading_finished (LoadingType.UPDATES);          
        }
        
        private void parse_error (GLib.Error error) {
            if ("Could not get lock" in error.message)
                error_received (ErrorType.COULD_NOT_GET_LOCK);
        }
        
        public void build_database () {
            database.build ();
        }
        
        public AppsManager (string db) {
            client = new Client();
            control = new Control();
            task = new Task();
            transaction = new Cancellable();
            
            database_path = db;
            
            try {
                database = new DBManager (database_path);
            } catch (GLib.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }

        }
    }
}
