/*
 * Copyright (c) 2012-2013 Mario Guerriero <mefrio.g@gmail.com>
 *
 * This is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

using PackageKit;
using SQLHeavy;
#if HAVE_GTK
using Gtk; // Needed for icontheme support
#endif

namespace AppStore {
    class AppInstallPackage {
        public string[] directories { get; private set; }
        public string name { get; private set; }
        public string icon { get; set; }
        public string summary { get; set; }
        public string id { get; private set; }
        
        internal AppInstallPackage (string[] directories, string name, string icon, string summary, string id) {
            this.directories = directories;
            this.name = name;
            this.icon = icon;
            this.summary = summary;
            this.id = id;
        }
    }

    class AppInstallDirectory {
        public string id { get; private set; }
        public string name { get; private set; }
        public string directory { get; private set; }
        
        internal AppInstallDirectory (string id, string name, string directory) {
            this.name = name;
            this.directory = directory;
        }
    }

    internal class DbBuilder {
        private Dir root;
        private KeyFile keys;
        private string desktop_dir;
        private MenuDir[] menu_dirs;
        private Gee.HashMap<string, AppInstallPackage> packages;
#if HAVE_GTK
        private IconTheme theme;
#endif
        private string category;
        
        private const int DEFAULT_MAX_ITEMS = 100;
        private GLib.KeyFile? pk_conf;

        private PackageKit.Task task;
        private PackageKit.Control control;
        private SQLHeavy.Transaction transaction;
        private SQLHeavy.Database db;
        
        void add_pkg (Package pk) {
            AppInstallPackage app_pkg = packages[pk.get_name()];
            if (app_pkg == null) { return; }
            if (app_pkg.summary == "") {
                app_pkg.summary = pk.get_summary();
            }
#if HAVE_GTK
            if (app_pkg.icon == "" || ! theme.has_icon (app_pkg.icon)) {
                app_pkg.icon = "applications-other";
            }
#else
            if (app_pkg.icon == "") {
                app_pkg.icon = "applications-other";
            }
#endif
            foreach (string dir in app_pkg.directories) {
                message_log ("Inserting " + app_pkg.id +" in " + dir, GLib.LogLevelFlags.LEVEL_DEBUG);
                Type t_string = typeof(string);
                try {
                    transaction.execute_insert("INSERT INTO '%s' VALUES (:id, :name, :description, :icon);".printf (dir),
                                             ":id", t_string, app_pkg.id,
                                             ":name", t_string, app_pkg.name,
                                             ":description", t_string, app_pkg.summary,
                                             ":icon", t_string, app_pkg.icon);
                } catch (SQLHeavy.Error e) {
                    message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
                }
            }
        }
        
        string[] get_included_in (string[] categories) {
            string[] result = {};
            foreach (MenuDir dir in menu_dirs) {
                if (dir.level != 1) { continue; }
                foreach (string cat in categories) {
                    if (cat in dir.excluded) {
                        result = {};
                        break;
                    }
                    if (cat in dir.included && ! (dir.id in result) ) {
                        result += dir.id;
                    }
                }
            }
            return result;
        }
        
        string get_safe_key (string val) {
            try {
                return keys.get_string ("Desktop Entry", val);
            } catch (GLib.Error e) {
                message_log (e.message);
            }
            return "";
        }
        
        void get_apps () {
            // Get apps from install-data dir
            root.rewind();
            string name;
            while ((name = root.read_name ()) != null) {
            if (! name.has_suffix (".desktop")) { continue; }
                try {
                    try {
                        keys.load_from_file (desktop_dir+name, 0);
                        if (keys.has_key("Desktop Entry", "Categories") == false) { continue; }
                        if (keys.has_key("Desktop Entry", "X-AppInstall-Ignore")) {
                            if (keys.get_boolean ("Desktop Entry", "X-AppInstall-Ignore")) {
                                message_log ("Ignoring " + name, GLib.LogLevelFlags.LEVEL_WARNING);
                                continue;
                            }
                        }
                        string[] categories = keys.get_string_list ("Desktop Entry", "Categories");
                        string[] tmp_dirs = get_included_in (categories);
                        AppInstallPackage pkg = new AppInstallPackage (tmp_dirs, get_safe_key ("Name"), get_safe_key ("Icon"), get_safe_key ("Comment"), get_safe_key ("X-AppInstall-Package"));
                        packages.set (pkg.id, pkg);
                    } catch (KeyFileError e) {
                        message_log (e.message);
                    }
                } catch (FileError e) {
                    message_log (e.message);
                }
            }
            
            // Now get all packages, also the ones which are not in app-install-data
            try {
                var result = task.get_packages (common_filter, null, () => {});
                result.get_package_array ().foreach ((p) => {
                    if (p == null)
                        return;
                    
                    // If it is not in app-install dir let's get it with PK
                    if (! packages.has_key (p.get_name ())) {
                        // Put them in the void fallback category
                        AppInstallPackage app = new AppInstallPackage ({""}, p.get_summary (), "deb", p.get_name (), p.get_name ());
                        packages.set (app.id, app); 
                    } 
                });
            } catch (GLib.Error e) {
                message_log (e.message);
            }
            
        }
        
        bool resolve_packages () {
            string[] ids = {};
            foreach (string s in packages.keys) {
                ids += s;
            }
            ids += null;
            
            try {
                transaction = db.begin_transaction ();
            } catch (SQLHeavy.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            
            // Just to check if the app can be resolved
            bool transaction_possible = true;
            
            // Max n items to parse to pk daemon per time
            int max_items = DEFAULT_MAX_ITEMS - 1;
            if (pk_conf != null) {
                try {     
                    max_items = pk_conf.get_integer ("Daemon", "MaximumItemsToResolve") - 1;
                } catch (KeyFileError e) {
                    max_items = DEFAULT_MAX_ITEMS - 1;
                }
            }
            
            try {
                for (int i=0; i<=ids.length; i+=max_items) {
                    string[] resolve_ids = ids[i:i+max_items];
                    resolve_ids += null;
                    Results result = task.resolve (common_filter, resolve_ids, null, () => {});
                    result.get_package_array().foreach ((pkg) => {
                        if (pkg != null) {
                            //var res = task.get_details_sync ({pkg.get_id (), null}, null, () => {});
                            add_pkg (pkg);
                        }
                    });
                }
            } catch (GLib.Error e) {
                message_log (e.message);
                transaction_possible = false;
            }
  
            try {
                if (transaction_possible)
                    transaction.commit();
            } catch (SQLHeavy.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            
            return true;
        }
        
        internal bool build () {
            desktop_dir = "/usr/share/app-install/desktop/";
            try {
                root = Dir.open (desktop_dir, 0);
            } catch (FileError e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            keys = new KeyFile ();
            
            MenuParser parser = new MenuParser ("/usr/share/app-install/desktop/applications.menu");
            
            try {
                transaction = db.begin_transaction ();
            } catch (SQLHeavy.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            
            message_log ("Collecting app-install data...", GLib.LogLevelFlags.LEVEL_DEBUG);
            
            try {
                transaction.execute ("CREATE TABLE 'DIRECTORIES' (id, name, summary, icon, directory);");
            } catch (SQLHeavy.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            
            menu_dirs = parser.parse();
            
            foreach (MenuDir item in menu_dirs) {
                if (item.level == 1) { // We just want the Applications childs
                    try {
                        transaction.execute ("CREATE TABLE '%s' (id, name, summary, icon);".printf (item.id));
                        transaction.execute ("INSERT INTO 'DIRECTORIES' VALUES (:id, :name, :summary, :icon, :directory);", 
                                       ":id", typeof (string), item.id,
                                       ":name", typeof (string), item.name,
                                       ":summary", typeof (string), item.summary,
                                       ":icon", typeof (string), item.icon,
                                       ":directory", typeof (string), item.directory);
                    } catch (SQLHeavy.Error e) {
                        message_log (e.message);
                    }
                }
            }
            // Insert a void dir too
            string void_dir = "";
            try {
                transaction.execute ("CREATE TABLE '%s' (id, name, summary, icon);".printf (void_dir));
                transaction.execute ("INSERT INTO 'DIRECTORIES' VALUES (:id, :name, :summary, :icon, :directory);", 
                              ":id", typeof (string), void_dir,
                              ":name", typeof (string), void_dir,
                              ":summary", typeof (string), void_dir,
                              ":icon", typeof (string), void_dir,
                              ":directory", typeof (string), void_dir);
            } catch (SQLHeavy.Error e) {
                message_log (e.message);
            }
            
            try {
                transaction.commit();
            } catch (SQLHeavy.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            
            get_apps ();
            
            message_log ("Querying packagekit...", GLib.LogLevelFlags.LEVEL_MESSAGE);
            
            return resolve_packages ();
        }
        
        internal bool build_fallback () {
            try {
                transaction = db.begin_transaction ();
            } catch (SQLHeavy.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            
            bool control_properties = true;
            
            try {
                control_properties = control.get_properties (null);
            } catch (GLib.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }            
            
            if (! control_properties) {
                return false;
            }
            string[] categs;
            categs = PackageKit.group_bitfield_to_string (control.groups).split(";");
            
            try {
                transaction.execute ("CREATE TABLE 'DIRECTORIES' (id, name, summary, icon, directory);");
            } catch (SQLHeavy.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            
            try {
                foreach (string c in categs) {
                    try {
                        transaction.execute ("CREATE TABLE '%s' (id, name, summary, icon);".printf (c));
                        transaction.execute ("INSERT INTO 'DIRECTORIES' VALUES (:id, :name, :summary, :icon, :directory);", 
                                             ":id", typeof (string), c,
                                             ":name", typeof (string), c,
                                             ":summary", typeof (string), c,
                                             ":icon", typeof (string), "applications-other",
                                             ":directory", typeof (string), "null");
                    } catch (SQLHeavy.Error e) {
                        message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
                    }
                    Results result = task.search_groups (Filter.APPLICATION | Filter.ARCH, {c, null}, null, () => {});
                    this.category = c; // closure support
                    result.get_package_array ().foreach ((pkg) => {
                        if (pkg != null) {
                            AppInstallPackage app = new AppInstallPackage ({category}, pkg.get_name(), "deb", pkg.get_summary(), pkg.get_name());
                            packages.set (app.id, app);
                            add_pkg (pkg);
                        }
                    });
                }
            } catch (GLib.Error e) {
                message_log (e.message);
                return false;
            }
            
            try {
                transaction.commit();
            } catch (SQLHeavy.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            
            return true;
        }
        
        internal DbBuilder (Database db) {
#if HAVE_GTK
            theme = new IconTheme();
            theme.append_search_path ("/usr/share/app-install/icons/");
#endif
            this.db = db;
            task = new Task ();
            control = new Control ();
            
            // PK conf file
            pk_conf = new KeyFile ();
            try {
                try {
                    pk_conf.load_from_file ("/etc/PackageKit/PackageKit.conf", GLib.KeyFileFlags.NONE);
                } catch (KeyFileError kfe) {
                    pk_conf = null;
                } 
            } catch (FileError fe) {
                pk_conf = null;
            }
            
            packages = new Gee.HashMap<string, AppInstallPackage> ();
        }
    }
    
    public class DBManager : GLib.Object {
        private Database database;
        private string database_path = null;
        
        /**
         * @param path the path where the application database is located.
         */
        public DBManager (string path) throws GLib.Error {
            this.database_path = path;

            bool exists = FileUtils.test (this.database_path, FileTest.EXISTS);
            
            if (exists)
                database = new Database (this.database_path, SQLHeavy.FileMode.READ);

        }
        
        /**
         * Build a new application database.
         * 
         * @param menu_filename the menu filename file.
         * @return true if succes, false otherwise.
         */
        public bool build () {
            if (FileUtils.test (this.database_path, FileTest.EXISTS))
                FileUtils.remove (this.database_path);
            
            try {   
                database = new Database (this.database_path);
            } catch (SQLHeavy.Error e) {
                message_log (e.message, GLib.LogLevelFlags.LEVEL_ERROR);
            }
            
            var dbbuilder = new DbBuilder (database);            
            
            if (FileUtils.test ("/usr/share/app-install", FileTest.EXISTS)) {
                return dbbuilder.build ();
            } else {
                return dbbuilder.build_fallback ();
            }

        }
        
        /**
         * Update application database.
         * 
         */
        public void update () {
        }
        
        /**
         * Get a QueryResult containing the packages rows
         * @param category_id the category id
         * @return a QueryResult
         */
        public QueryResult get_apps_for_category (string category_id) throws GLib.Error {
            Query get_query = database.prepare ("SELECT * FROM '%s';".printf (category_id));
            return get_query.execute ();
        }
        
        /**
         * Get the categories rows
         * @return a QueryResult
         */
        public QueryResult get_categories () throws GLib.Error {
            Query cat_query = database.prepare ("SELECT * FROM 'DIRECTORIES';");
            return cat_query.execute ();
        }
        
        /**
         * Get the number of rows in a table
         * @param category_id the category id
         * @return a int representing the number of apps
         */
        public int get_rows_for_category (string category_id) throws GLib.Error {
            Query q = database.prepare ("SELECT COUNT (*) FROM '%s';".printf (category_id));
            QueryResult r = q.execute();
            return r.fetch_int (0);
        }
    }
}
